#include <iostream>
#include <memory>

// usate 'c++filt -t' nella linea di comando
// per ottenere il nome completo dei tipi

int* usa_puntatore()
{
  int *pi1{new int(4)};

  std::cout << "*pi1: " << *pi1 << std::endl; 

  *pi1 = *pi1 + 3;

  std::cout << "*pi1: " << *pi1 << std::endl; 

  int *pi2;

  pi2 = pi1;

  *pi2 = 6;

  std::cout << "*pi1: " << *pi1 << std::endl; 

  return pi2;
}

void puntatori_tradizionali()
{
  std::cout << "PUNTATORI TRADIZIONALI" << std::endl;

  auto pi{usa_puntatore()};

  std::cout << "*pi: " << *pi << std::endl;

  // qui genero un memory leak perché l'area di memoria 
  // puntata da 'pi' *non* viene mai de-allocata e al 
  // termine della funzione ogni puntatore ad essa viene
  // perduto

  // delete pi;    // decommentando questa linea di codice
                   // è possibile rimuovere il memory leak
}

std::shared_ptr<int> usa_shared_pointer()
{
  std::shared_ptr<int> s1{new int(4)};

  std::cout << "*s1: " << *s1 << std::endl; 

  *s1 = *s1 + 3;    // posso usare lo smart pointer con la 
                    // stessa sintassi dei puntatori

  std::cout << "*s1: " << *s1 << std::endl;
  
  std::cout << "s1.use_count(): " 
            << s1.use_count()    // riporta il numero di shared pointer
            << std::endl;      // che si riferiscono alla stessa area 
                                 // di memoria

  std::shared_ptr<int> s2;  // gli shared pointer possono 
                            // essere dichiarati senza 
                            // inizializzarli

  s2 = s1;          // gli shared pointer sono copiabili:
                    // l'area di memoria a cui si riferiscono
                    // è la stessa

  std::cout << "s1.use_count(): " 
            << s1.use_count()    
            << std::endl;  

  *s2 = 6;

  std::cout << "*s1: " << *s1 << std::endl;

  return s2;  // quando il valore in s2 viene restituito 
              // s1 e s2 smettono di esistere, ma esiste 
              // ancora un riferimento all'area di memoria
              // a cui puntavano: il valore restituito
}

void shared_pointer()
{
  std::cout << "SHARED POINTER" << std::endl;

  auto sp{usa_shared_pointer()};

  std::cout << "*sp: " << *sp << std::endl;

  std::cout << "sp.use_count(): " 
            << sp.use_count()    
            << std::endl;  
}

std::shared_ptr<int> usa_unique_pointer()
{
  std::unique_ptr<int> u1{new int(4)};

  std::cout << "*u1: " << *u1 << std::endl; 

  *u1 = *u1 + 3;    // posso usare lo smart pointer con la 
                    // stessa sintassi dei puntatori

  std::cout << "*u1: " << *u1 << std::endl;

  std::unique_ptr<int> u2;  // gli unique pointer possono 
                            // essere dichiarati senza 
                            // inizializzarli

  // u2 = u1;          // gli unique pointer *non* sono copiabili:
                       // l'area di memoria a cui si riferiscono
                       // è la stessa

  return u1;  // quando il valore in u1 viene restituito 
              // u1 smette di esistere, ma esiste 
              // ancora un riferimento all'area di memoria
              // a cui puntava: il valore restituito
}

void unique_pointer()
{
  std::cout << "UNIQUE POINTER" << std::endl;
 
  auto up{usa_unique_pointer()};

  std::cout << "*up: " << *up << std::endl;
}

int main()
{
  puntatori_tradizionali();
  unique_pointer();
  shared_pointer();

  return 0;
}