#include <iostream>
#include <memory>

bool is_prime(const std::unique_ptr<unsigned int[]>& P, const size_t& P_size,
              const unsigned int& value)
{
  // P contiene tutti i primi minori di value in ordine
  
  // per tutti i primi tali che il loro quadrato è minore di value
  for (size_t j=0; j<P_size && P[j]*P[j]<=value; ++j) {

    // se value è divisibile interamente dal primo 
    if (value%P[j]==0) {

      // value non è un primo
      return false;
    } 
  } 


  // se tutti i primi minori della radice quadrata di value non 
  // dividono value, value è primo
  return true;
}

unsigned int Eratosthenes(const unsigned int& n)
{
  // c'è un errore per n==0

  // alloco un array per memorizzare i primi n primi
  std::unique_ptr<unsigned int[]> P{new unsigned int[n]};

  P[0] = 2;  // il primo primo è 2

  size_t i{1};  // il numero di primi già scoperti è 1

  unsigned int value{P[0]+1};  // itero su tutti i naturali usando
                               // 'value' come variabile
  while (i<n) {
    if (is_prime(P, i, value)) {
      P[i++] = value;
    }

    ++value;
  }

  return P[n-1];
}

int main()
{
  constexpr size_t n{10};  

  std::cout << "Eratosthenes(" << n << ")= " << Eratosthenes(n) << std::endl;

  return 0;
}